import backoff
import openai

from litellm import acompletion
from typing import List, Dict
from fastinference.data_processing.datablock import DataBlock


class LLMManager:
    def __init__(self, **kwargs) -> None:
        self.__dict__.update(kwargs)

    @backoff.on_exception(backoff.expo,
                          (openai.APITimeoutError,
                           openai.BadRequestError,
                           openai.AuthenticationError,
                           openai.PermissionDeniedError,
                           openai.NotFoundError,
                           openai.UnprocessableEntityError,
                           openai.RateLimitError),
                          max_time=300)
    async def acompletions_with_backoff(self, data: DataBlock, **kwargs):
        extract_content_with_prompt = data.content_with_prompt.render_prompt_for_litellm()
        data.response = await acompletion(**self.__dict__, **kwargs, messages=extract_content_with_prompt)
        return data

    async def get_acompletion(self, content: DataBlock):
        return await self.acompletions_with_backoff(data=content)
        



from dataclasses import dataclass

from fastinference.prompt.prompt_template import PromptTemplate


@dataclass
class DataBlock:
    content: str
    metadata: dict = None
    content_with_prompt: PromptTemplate = None

    def __post_init__(self):
        if self.metadata is None:
            self.metadata = {}

    def is_time_series(self) -> bool:
        return 'date_stamp' in self.metadata



from fastinference.managers.task_manager import TasksManager
from fastinference.data_processing.datablock import DataBlock

from typing import List, Tuple, Dict
from itertools import chain

import pandas as pd
import numpy as np
import os


def extract_response_only(data: List[DataBlock], task_manager: TasksManager) -> List:
    def get_response(data):
        return data.response.choices[0].message.content

    return task_manager.build_multithread(get_response, data, "Building the LLM Response only")


def extract_from_file(path: str,
                      column_main_content: str,
                      task_manager: TasksManager,
                      nb_chunks: int = 20) -> List[Tuple[str, Dict]]:
    def extract_chunks(data_chunk: pd.DataFrame, column_main_content: str) -> List[Tuple[str, Dict]]:
        result = []
        for index, row in data_chunk.iterrows():
            main_content = row[column_main_content]
            other_data = {col: row[col] for col in data_chunk.columns if col != column_main_content}
            result.append((main_content, other_data))
        return result

    # Identify the file extension
    _, file_extension = os.path.splitext(path)
    file_extension = file_extension.lower()

    # Read data based on file extension
    if file_extension == ".csv":
        data = pd.read_csv(path)
    elif file_extension == ".xlsx":
        data = pd.read_excel(path)
    elif file_extension == ".json":
        data = pd.read_json(path)
    elif file_extension == ".parquet":
        data = pd.read_parquet(path)
    else:
        raise ValueError("Unsupported file type")

    # Split data into chunks
    data_chunks = np.array_split(data, nb_chunks)

    # Use the task manager to process data chunks in parallel
    list_formatted = task_manager.build_multithread(extract_chunks,
                                                    data_chunks,
                                                    f"Formating the {file_extension} file",
                                                    column_main_content=column_main_content)

    # Flatten the list of results from all chunks
    list_flatten = list(chain.from_iterable(list_formatted))

    return list_flatten




from typing import Tuple, Dict, List
from fastinference.prompt.prompt_template import PromptTemplate
from fastinference.data_processing.datablock import DataBlock


class DataProcessor:
    def __init__(self, raw_data: List[Tuple[str, Dict]], task_manager, prompt: str = None) -> None:
        self.raw_data = raw_data
        self.task_manager = task_manager
        self.prompt = prompt
        self.datablock_chain = self.build_data_chain()

    def build_data_chain(self) -> List[DataBlock]:
        return self.task_manager.build_multithread(self.build_data_block, self.raw_data, "Building Data Chain")

    def build_data_block(self, data: Tuple[str, Dict]):
        if data[1] is not None:
            datablock = DataBlock(content=data[0],
                                  metadata=data[1])

            if self.prompt:
                datablock.content_with_prompt = PromptTemplate(datablock, self.prompt)
        else:
            datablock = DataBlock(content=data[0])

        if self.prompt:
            datablock.content_with_prompt = PromptTemplate(datablock, self.prompt)

        return datablock

    @staticmethod
    def build_litellm_prompt_in_datablock(data: DataBlock) -> DataBlock:
        try:
            litellm_prompt = data.prompt_with_content.render_prompt_for_litellm()
        except Exception as e:
            raise Exception('Could not get the prompt_with_content attribute in the PromptTemplate class: ', e)
        data.__setattr__('litellm_prompt_object', litellm_prompt)
        return data

    def render_prompt_for_many(self, prompt_core: str) -> None:
        try:
            self.task_manager.build_multithread(self.build_prompt_in_datablock, self.datablock_chain,
                                                task_name="Building Prompt in DataChain", prompt_core=prompt_core)
        except Exception as e:
            raise Exception('Could not build the PromptTemplate in the DataBlock object: ', e)

    def render_prompt_for_many_litellm_format(self) -> None:
        try:
            self.task_manager.build_multithread(self.build_litellm_prompt_in_datablock, self.datablock_chain,
                                                task_name="Building LiteLLM object in DataBlockChain")
        except Exception as e:
            raise Exception('Could not build the LiteLLM prompt format in the DataBlockChain: ', e)



import asyncio

from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import partial

from tqdm import tqdm
from typing import List, Callable

from fastinference.data_processing.datablock import DataBlock


class TasksManager:
    def __init__(self, nb_thread: int = None) -> None:
        self.nb_thread = nb_thread

    def build_multithread(self, function_to_execute: Callable, data: list, task_name: str, **kwargs) -> List:
        tasks = []

        func_with_kwargs = partial(function_to_execute, **kwargs)
        with ThreadPoolExecutor(max_workers=self.nb_thread) as executor:
            future_to_date = {executor.submit(func_with_kwargs, content) for content in data}
            for future in tqdm(as_completed(future_to_date), total=len(future_to_date), desc=task_name):
                result = future.result()
                if result is not None:
                    tasks.append(result)

        return tasks

    async def build_async(self,
                          function_to_execute: Callable,
                          datablock_chain: List[DataBlock],
                          **kwargs) -> List[asyncio.Task]:

        loop = asyncio.get_running_loop()
        with ThreadPoolExecutor(max_workers=self.nb_thread) as executor:
            futures = [executor.submit(self.prepare_task, loop, function_to_execute, block, **kwargs)
                       for block in datablock_chain]
            tasks = [future.result() for future in futures]

        return tasks

    def prepare_task(self, loop, function_to_execute, prompt_object, **kwargs):
        return loop.create_task(function_to_execute(prompt_object, **kwargs))




from typing import List, Dict


class PromptTemplate:
    def __init__(self, data, core_prompt: str):
        self.data = data
        self.core_prompt = core_prompt
        self.prompt = self.render_prompt()

    def render_prompt(self) -> str:
        try:
            format_dict = {'content': self.data.content, **self.data.metadata}
            return self.core_prompt.format(**format_dict)
        except KeyError as e:
            raise ValueError(f"Missing variable {e} in the prompt template.")

    def render_prompt_for_litellm(self) -> List[Dict]:
        try:
            return [{"content": self.prompt, "role": "user"}]
        except Exception as e:
            raise Exception(f'Could not convert the prompt in the litellm format: {e}')


from distutils.core import setup
setup(
  name = 'fastinference-llm',
  packages=[
    'fastinference',
    'fastinference.data_processing',
    'fastinference.managers',
    'fastinference.prompt',
    'fastinference.utils',
  ],
  version = '0.0.4',
  license='MIT',
  description = 'Seamlessly integrate with top LLM APIs for speedy, robust, and scalable querying. Ideal for developers needing quick, reliable AI-powered responses.',   # Give a short description about your library
  author = 'Baptiste Lefort',
  author_email = 'lefort.baptiste@icloud.com',
  url = 'https://github.com/blefo/FastInference',
  keywords = ['api', 'fast', 'inference', 'distributed', 'llm'],
  install_requires=[
        'pandas',
        'numpy',
        'backoff',
        'openai',
        'litellm',
        'tqdm',
      ],
  classifiers=[
    'Programming Language :: Python :: 3',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
  ],
  long_description=open('README.md').read(),
  long_description_content_type='text/markdown',
)



from fastinference.managers.llm_manager import LLMManager
from fastinference.managers.task_manager import TasksManager
from fastinference.data_processing.data_processor import DataProcessor
from fastinference.utils.data_processing import extract_from_file, extract_response_only

import asyncio


class FastInference:
    def __init__(self, file_path: str, main_column: str, prompt: str, only_response: bool = False, **kwargs):
        self.llm_manager = LLMManager(**kwargs)
        self.task_manager = TasksManager()
        self.file_path = file_path
        self.main_column = main_column
        self.prompt = prompt
        self.only_response = only_response

    def run(self):
        return asyncio.run(self.run_process())

    async def run_process(self):
        # Build the DataBlockChain
        data_loaded = extract_from_file(self.file_path,
                                        self.main_column,
                                        self.task_manager)[:200]

        data = DataProcessor(data_loaded, self.task_manager, prompt=self.prompt)

        # Run the LLM
        tasks = await self.task_manager.build_async(self.llm_manager.get_acompletion, data.datablock_chain)
        results = await asyncio.gather(*tasks)

        if self.only_response:
            return extract_response_only(results, self.task_manager)
        else:
            return results